@@ -1,5 +1,20 @@
+# Copyright 2016-2018 The NATS Authors
+# Licensed under the Apache License, Version 2.0 (the "License");
+# you may not use this file except in compliance with the License.
+# You may obtain a copy of the License at
+#
+# http://www.apache.org/licenses/LICENSE-2.0
+#
+# Unless required by applicable law or agreed to in writing, software
+# distributed under the License is distributed on an "AS IS" BASIS,
+# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
+# See the License for the specific language governing permissions and
+# limitations under the License.
+#
+
 require 'nats/io/parser'
 require 'nats/io/version'
+require 'nats/nuid'
 require 'thread'
 require 'socket'
 require 'json'
@@ -159,13 +174,20 @@ module NATS
         # Hostname of current server; used for when TLS host
         # verification is enabled.
         @hostname = nil
+
+        # New style request/response implementation.
+        @resp_sub = nil
+        @resp_map = nil
+        @resp_sub_prefix = nil
+        @nuid = NATS::NUID.new
       end
 
-      # Establishes connection to NATS
+      # Establishes connection to NATS.
       def connect(opts={})
         opts[:verbose] = false if opts[:verbose].nil?
         opts[:pedantic] = false if opts[:pedantic].nil?
         opts[:reconnect] = true if opts[:reconnect].nil?
+        opts[:old_style_request] = false if opts[:old_style_request].nil?
         opts[:reconnect_time_wait] = RECONNECT_TIME_WAIT if opts[:reconnect_time_wait].nil?
         opts[:max_reconnect_attempts] = MAX_RECONNECT_ATTEMPTS if opts[:max_reconnect_attempts].nil?
         opts[:ping_interval] = DEFAULT_PING_INTERVAL if opts[:ping_interval].nil?
@@ -198,6 +220,12 @@ module NATS
           }
         end
 
+        if @options[:old_style_request]
+          # Replace for this instance the implementation
+          # of request to use the old_request style.
+          class << self; alias_method :request, :old_request; end
+        end
+
         # Check for TLS usage
         @tls = @options[:tls]
 
@@ -344,11 +372,58 @@ module NATS
         sid
       end
 
-      # Sends a request expecting a single response or raises a timeout
-      # in case the request is not retrieved within the specified deadline.
+      # Sends a request using expecting a single response using a
+      # single subscription per connection for receiving the responses.
+      # It times out in case the request is not retrieved within the
+      # specified deadline.
       # If given a callback, then the request happens asynchronously.
       def request(subject, payload, opts={}, &blk)
         return unless subject
+
+        # If a block was given then fallback to method using auto unsubscribe.
+        return old_request(subject, payload, opts, &blk) if blk
+
+        token = nil
+        inbox = nil
+        future = nil
+        response = nil
+        timeout = opts[:timeout] ||= 0.5
+        synchronize do
+          start_resp_mux_sub! unless @resp_sub_prefix
+
+          # Create token for this request.
+          token = @nuid.next
+          inbox = "#{@resp_sub_prefix}.#{token}"
+
+          # Create the a future for the request that will
+          # get signaled when it receives the request.
+          future = @resp_sub.new_cond
+          @resp_map[token][:future] = future
+        end
+
+        # Publish request and wait for reply.
+        publish(subject, payload, inbox)
+        with_nats_timeout(timeout) do
+          @resp_sub.synchronize do
+            future.wait(timeout)
+          end
+        end
+
+        # Check if there is a response already
+        synchronize do
+          result = @resp_map[token]
+          response = result[:response]
+        end
+
+        response
+      end
+
+      # Sends a request creating an ephemeral subscription for the request,
+      # expecting a single response or raising a timeout in case the request
+      # is not retrieved within the specified deadline.
+      # If given a callback, then the request happens asynchronously.
+      def old_request(subject, payload, opts={}, &blk)
+        return unless subject
         inbox = new_inbox
 
         # If a callback was passed, then have it process
@@ -520,7 +595,7 @@ module NATS
             future.signal
 
             return
-          elsif sub.callback
+          elsif sub.pending_queue
             # Async subscribers use a sized queue for processing
             # and should be able to consume messages in parallel.
             if sub.pending_queue.size >= sub.pending_msgs_limit \
@@ -1093,6 +1168,47 @@ module NATS
         @ping_interval_thread.abort_on_exception = true
       end
 
+      # Prepares requests subscription that handles the responses
+      # for the new style request response.
+      def start_resp_mux_sub!
+        @resp_sub_prefix = "_INBOX.#{@nuid.next}"
+        @resp_map = Hash.new { |h,k| h[k] = { }}
+
+        @resp_sub = Subscription.new
+        @resp_sub.subject = "#{@resp_sub_prefix}.*"
+        @resp_sub.received = 0
+
+        # FIXME: Allow setting pending limits for responses mux subscription.
+        @resp_sub.pending_msgs_limit = DEFAULT_SUB_PENDING_MSGS_LIMIT
+        @resp_sub.pending_bytes_limit = DEFAULT_SUB_PENDING_BYTES_LIMIT
+        @resp_sub.pending_queue = SizedQueue.new(@resp_sub.pending_msgs_limit)
+        @resp_sub.wait_for_msgs_t = Thread.new do
+          loop do
+            msg = @resp_sub.pending_queue.pop
+            @resp_sub.pending_size -= msg.data.size
+
+            # Pick the token and signal the request under the mutex
+            # from the subscription itself.
+            token = msg.subject.split('.').last
+            future = nil
+            synchronize do
+              future = @resp_map[token][:future]
+              @resp_map[token][:response] = msg
+            end
+
+            # Signal back that the response has arrived.
+            @resp_sub.synchronize do
+              future.signal
+            end
+          end
+        end
+
+        sid = (@ssid += 1)
+        @subs[sid] = @resp_sub
+        send_command("SUB #{@resp_sub.subject} #{sid}#{CR_LF}")
+        @flush_queue << :sub
+      end
+
       def can_reuse_server?(server)
         return false if server.nil?
 
